import { objectSpread2 as _objectSpread2 } from '../_virtual/_rollupPluginBabelHelpers.js';
import { combineReducers } from 'redux';
import { schema2node, schema2tree, getNodeByRjsfId, updateNode, moveNode, removeNode, addNode } from './core.js';
import defaultSettings from './default/settings.js';
import defaultMenuSchema from './default/menu.js';
import undoable, { includeAction } from 'redux-undo';

var defaultMenu = schema2node(['menu'], defaultMenuSchema.schema, defaultMenuSchema.uiSchema);
var DEFAULT_TREE_NAME = 'root';
var emptyTree = schema2tree(DEFAULT_TREE_NAME, {
  type: 'object',
  properties: {}
});

function tree() {
  var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : emptyTree;
  var action = arguments.length > 1 ? arguments[1] : undefined;

  switch (action.type) {
    case 'TREE_CLEAR':
      return emptyTree;

    case 'TREE_SET_TREE':
      {
        var _action$payload = action.payload,
            name = _action$payload.name,
            schema = _action$payload.schema,
            uiSchema = _action$payload.uiSchema;
        return schema2tree(name || state[0] && state[0].name || DEFAULT_TREE_NAME, schema || state[0].schema, uiSchema || state[0].uiSchema);
      }

    case 'TREE_ADD_NODE':
      {
        var _action$payload2 = action.payload,
            targetNodeKey = _action$payload2.targetNodeKey,
            position = _action$payload2.position,
            node2add = _action$payload2.node2add;
        return addNode(state, targetNodeKey, position, node2add);
      }

    case 'TREE_REMOVE_NODE':
      var newState = removeNode(state, action.payload);
      return newState.length ? newState : emptyTree;

    case 'TREE_MOVE_NODE':
      {
        var _action$payload3 = action.payload,
            source = _action$payload3.source,
            target = _action$payload3.target,
            _position = _action$payload3.position;
        return moveNode(state, source, target, _position);
      }

    case 'TREE_UPDATE_NODE':
      {
        var _action$payload4 = action.payload,
            _target = _action$payload4.target,
            nodeUpdate = _action$payload4.nodeUpdate;

        var _newState = updateNode(state, _target, nodeUpdate);

        return _newState;
      }

    default:
      return state;
  }
}

function activeNodeKey() {
  var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;
  var action = arguments.length > 1 ? arguments[1] : undefined;

  switch (action.type) {
    case 'ACTIVE_NODE_KEY_SET':
      return action.payload || null;

    case 'TREE_UPDATE_NODE':
      var _action$payload5 = action.payload,
          target = _action$payload5.target,
          nodeUpdate = _action$payload5.nodeUpdate;

      if (state === target && nodeUpdate.name) {
        var path = state.split('.');

        if (path[path.length - 1] !== nodeUpdate.name) {
          return path.slice(0, -1).join('.') + '.' + nodeUpdate.name;
        }
      }

      return state;

    case 'TREE_SET_TREE':
    case 'TREE_CLEAR':
      return null;

    default:
      return state;
  }
}

function formData() {
  var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
  var action = arguments.length > 1 ? arguments[1] : undefined;

  switch (action.type) {
    case 'FORM_DATA_SET':
      return action.payload;

    default:
      return state;
  }
}

function settings() {
  var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : defaultSettings;
  var action = arguments.length > 1 ? arguments[1] : undefined;

  switch (action.type) {
    case 'SETTINGS_UPDATE':
      return Object.assign({}, state, action.payload);

    default:
      return state;
  }
}

function menu() {
  var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : defaultMenu;
  var action = arguments.length > 1 ? arguments[1] : undefined;

  switch (action.type) {
    case 'MENU_SET':
      {
        var _action$payload6 = action.payload,
            schema = _action$payload6.schema,
            uiSchema = _action$payload6.uiSchema;
        return schema2node(['menu'], schema, uiSchema);
      }

    case 'MENU_DEFAULT':
      {
        return defaultMenu;
      }

    default:
      return state;
  }
}

function menuOpenKeys() {
  var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];
  var action = arguments.length > 1 ? arguments[1] : undefined;

  switch (action.type) {
    case 'MENU_OPEN_KEYS_SET':
      return action.payload;

    default:
      return state;
  }
}

function removeUnnecessaryHistory(reducer) {
  return function (state, action) {
    var newState = reducer(state, action);
    delete newState.history;
    return newState;
  };
}

var reducer = combineReducers({
  tree: removeUnnecessaryHistory(undoable(tree, {
    filter: includeAction(['TREE_UPDATE_NODE', 'TREE_SET_TREE', 'TREE_CLEAR', 'TREE_ADD_NODE', 'TREE_REMOVE_NODE', 'TREE_MOVE_NODE'])
  })),
  activeNodeKey: activeNodeKey,
  settings: settings,
  menu: menu,
  formData: formData,
  menuOpenKeys: menuOpenKeys
});
function reducer$1 (state, action) {
  console.log(state, action);

  switch (action.type) {
    case 'TREE_REMOVE_NODE_BY_RJSF_ID':
      {
        var target = getNodeByRjsfId(state.tree.present, action.payload);
        if (!target) return state;
        action = {
          type: 'TREE_REMOVE_NODE',
          payload: target.key
        };
        break;
      }

    case 'ACTIVE_NODE_KEY_SET_BY_RJSF_ID':
      {
        var _target2 = getNodeByRjsfId(state.tree.present, action.payload);

        if (!_target2) return state;
        action = {
          type: 'ACTIVE_NODE_KEY_SET',
          payload: _target2.key
        };
        break;
      }

    case 'TREE_UPDATE_NODE_BY_RJSF_ID':
      {
        var rjsfId = action.payload.rjsfId;

        var _target3 = getNodeByRjsfId(state.tree.present, rjsfId);

        if (!_target3) return state;
        action = {
          type: 'TREE_UPDATE_NODE',
          payload: _objectSpread2({}, action.payload, {
            target: _target3.key
          })
        };
        break;
      }
  }

  return reducer(state, action);
}

export default reducer$1;
